#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <err.h>
#include <fcntl.h>
#include <limits.h>
#include <fts.h>

#include "util.h"

#define LEN(s) (sizeof(s) / sizeof(s[0]))
#define POLLS_DIR "polls"

static char fname[PATH_MAX];
static char poll[1024];
static char create[2];
static char question[4096];
static char options[4096];
static char choice[16];

void
http_status(int statuscode)
{
	switch(statuscode) {
	case 307:
		printf("Status: 307 Temporary Redirect\r\n");
		printf("Location: /vote?poll=%s&choice=0\r\n\r\n", poll);
		break;
	case 401:
		printf("Status: 401 Bad Request\r\n\r\n");
		break;
	case 404:
		printf("Status: 404 Not Found\r\n\r\n");
		break;
	case 500:
		printf("Status: 500 Internal Server Error\r\n\r\n");
		break;
	default:
		printf("Status: 500 Internal Server Error\r\n\r\n");
		err(1, "unknown status code %d\n", statuscode);
	}
}

char *
pollfile(const char *poll_name, const char *postfix)
{
	char buf[PATH_MAX];
	int r;

	strlcpy(buf, poll_name, sizeof(buf));
	escapechars(buf);
	r = snprintf(fname, sizeof(fname), "%s/%s%s",
		POLLS_DIR, buf, postfix);
	if (r < 0 || (size_t)r >= sizeof(fname)) {
		http_status(500);
		err(1, "show_poll: snprintf fname %s/%s%s",
			POLLS_DIR, buf, postfix);
	}

	return fname;
}

void
print_html_head(void)
{
	printf("Content-type: text/html; charset=utf-8\r\n\r\n");
	puts("<!DOCTYPE html>\n"
		"<html>\n"
		"<head>\n"
		"	<style type=\"text/css\">\n"
		"		body {\n"
		"			margin: 1em auto;\n"
		"			max-width: 40em;\n"
		"			padding: 0 .62em;\n"
		"			font: 16px/1.62 sans-serif;\n"
		"			line-height: 1.3;\n"
		"		}\n"
		"		h1, h2, h3 {\n"
		"			line-height: 1.2;\n"
		"		}\n"
		"		a {\n"
		"			color: #000;\n"
		"		}\n"
		"		td.choice {\n"
		"			padding: .5em;\n"
		"		}\n"
		"		table.create, td.input, input.name,\n"
		"		textarea.question, textarea.options {\n"
		"			box-sizing: border-box;\n"
		"			width: 100%;\n"
		"		}\n"
		"	</style>\n"
		"</head>\n"
		"<body>\n");
}

void
print_html_foot(void)
{
	printf("</body>\n"
		"</html>\n");
}

int
print_poll_line(char *line, size_t *i, int intable, int vote)
{
	size_t c;

	if (sscanf(line, "%ld\t%4095[^\n]", &c, options) == 2) {
		if (!intable) {
			puts("</p>");
			if (vote) {
				puts("<form method=\"get\" action=\"\">");
				printf("<input type=\"hidden\" name=\"poll\" "
					"value=\"%s\" />\n", poll);
			}
			puts("<table>");
		}
		if (vote) {
			(*i)++;
			printf("\t<tr><td class=\"choice\">");
			printf("<input type=\"radio\" "
				"id=\"o%ld\" name=\"choice\" value=\"%ld\" />",
				*i, *i);
			printf("</td><td><label for=\"o%ld\">%s</label></td></tr>\n",
				*i, options);
		} else
			printf("\t<tr><td class=\"choice\">%ld</td><td>%s</td></tr>\n",
				c, options);
		return 1;
	} else {
		printf("%s<br/>\n", line);
		return 0;
	}
}

void
print_poll_file(FILE *fp, int vote)
{
	char *line = NULL;
	size_t linesize = 0, lineno = 0, i = 0;
	ssize_t linelen;
	int intable = 0;

	while ((linelen = getline(&line, &linesize, fp)) != -1) {
		lineno++;
		if (line[linelen - 1] == '\n')
			line[--linelen] = '\0';

		if (lineno == 1) {
			printf("<h2>");
			fwrite(line, linelen, 1, stdout);
			printf("</h2>\n<p>");
		} else {
			intable = print_poll_line(line, &i, intable, vote);
		}
	}
	free(line);
	if (ferror(fp)) {
		http_status(500);
		err(1, "print_poll_file: getline");
	}

	puts("</table>");
	if (vote) {
		puts("<input type=\"submit\" value=\"Submit\" "
			"class=\"button\"/>");
		puts("</form>");
		/* printf("<p><a href=\"?poll=%s&choice=0\">"
			"Show results</a></p>\n", poll); */
	}
}

int
create_poll_file(const char *name, const char *question, const char *options)
{
	FILE *fp;
	struct stat sb;
	size_t col;
	char *fname;

	if (!*name || !*question || !*options) {
		http_status(401);
		puts("<p><b>Error: Could not create poll</b></p>");
		puts("<ul>");
		if (!*name)
			puts("<li>Poll name is missing</li>");
		if (!*question)
			puts("<li>Poll question is missing</li>");
		if (!*options)
			puts("<li>Poll options are missing</li>");
		puts("</ul>");
		return -1;
	}

	fname = pollfile(name, "");
	if (stat(fname, &sb) == 0) {
		http_status(401);
		printf("<p>Poll '%s' already exists</p>", name);
		return -1;
	} else {
		if (!(fp = fopen(fname, "w"))) {
			http_status(404);
			exit(1);
		} else {
			fputs(question, fp);
			fputc('\n', fp);

			for (col = 0; *options; (void)*options++) {
				if (++col == 1 && *options != '\n' && *options != '\r')
					fputs("0\t", fp);

				switch(*options) {
				case '\t':
					fputc(' ', fp);
					break;
				case '\r':
					break;
				case '\n':
					if (col < 3) {
						col = 0;
						break;
					}
					col = 0;
				default:
					fputc(*options, fp);
					break;
				}
			}
			fputc('\n', fp);
			fclose(fp);
		}
	}
	return 0;
}

void
show_poll(const char *poll, int vote)
{
	FILE *fp;

	if (pledge("stdio rpath", NULL) == -1) {
		http_status(500);
		err(1, "show_poll: pledge");
	}

	if (!(fp = fopen(pollfile(poll, ""), "r"))) {
		http_status(404);
		exit(1);
	} else {
		print_poll_file(fp, vote);
		fclose(fp);
	}
}

void
list_polls(void)
{
	FTS *ftsp;
	FTSENT *p;
	int fts_options = FTS_COMFOLLOW | FTS_LOGICAL | FTS_NOCHDIR;
	char *paths[] = { (char*)POLLS_DIR, NULL };

	if (pledge("stdio rpath", NULL) == -1) {
		http_status(500);
		err(1, "list_polls: pledge");
	}

	if ((ftsp = fts_open(paths, fts_options, NULL)) == NULL) {
		http_status(500);
		err(1, "list_polls: fts_open");
	}

	puts("<h2>Poll listing</h2>");
	puts("<ul>");

	while ((p = fts_read(ftsp)) != NULL) {
		switch (p->fts_info) {
		case FTS_F:
			printf("<li><a href=\"?poll=%s\">%s</a></li>\n",
			       p->fts_path + LEN(POLLS_DIR),
			       p->fts_path + LEN(POLLS_DIR));
			break;
		default:
			break;
		}
	}
	fts_close(ftsp);
	puts("</ul>");
}

void
increment_option(char *poll, size_t n)
{
	static char fname_tmp[PATH_MAX];
	FILE *fp, *fp_tmp;
	size_t v, lineno = 0;
	char *line = NULL, *fname = NULL;
	size_t linesize = 0;
	ssize_t linelen;
	struct stat sb;

	fname = pollfile(poll, "_lock");
	strlcpy(fname_tmp, fname, sizeof(fname_tmp));
	while (stat(fname_tmp, &sb) == 0)
		usleep(100);
	if (!(fp_tmp = fopen(fname_tmp, "w"))) {
		http_status(500);
		err(1, "increment_option: fopen fp_tmp");
	}

	fname = pollfile(poll, "");
	if (!(fp = fopen(fname, "r"))) {
		http_status(404);
		err(1, "increment_option: fopen fp");
	}

	while ((linelen = getline(&line, &linesize, fp)) != -1) {
		if (sscanf(line, "%ld\t%4095[^\n]", &v, options) != 2)
			fputs(line, fp_tmp);
		else {
			if (++lineno == n)
				v++;
			fprintf(fp_tmp, "%zu\t%s\n", v, options);
		}
	}
	
	free(line);
	if (ferror(fp) || ferror(fp_tmp)) {
		http_status(500);
		err(1, "increment_option: getline");
	}
	fclose(fp);
	fclose(fp_tmp);

	if (rename(fname_tmp, fname) != 0) {
		http_status(500);
		err(1, "increment_option: rename");
	}
}

void
print_poll_create_form(void)
{
	puts("<h2>Create new poll</h2>");
	puts("<form method=\"get\" action=\"\">\n"
		"<input type=\"hidden\" name=\"create\" value=\"1\" />\n"
		"<table class=\"create\" width=\"100%\" border=\"0\" "
		"cellpadding=\"0\" cellspacing=\"0\">\n"
		"<tr>\n"
		"	<td width=\"100%\" class=\"input\">\n"
		"		<input type=\"text\" name=\"poll\" "
		"placeholder=\"Name\" class=\"name\" required />\n"
		"	</td>\n"
		"</tr>\n"
		"<tr>\n"
		"	<td width=\"100%\" class=\"input\">\n"
		"		<textarea rows=\"2\" name=\"question\" "
		"placeholder=\"Question (first line is header)\" "
		"class=\"question\" required></textarea>\n"
		"	</td>\n"
		"</tr>\n"
		"<tr>\n"
		"	<td width=\"100%\" class=\"input\">\n"
		"       <textarea rows=\"5\" name=\"options\" "
		"placeholder=\"Options (one per line)\" class=\"options\" required></textarea>\n"
		"	</td>\n"
		"</tr>\n"
		"<tr>\n"
		"	<td nowrap class=\"nowrap\">\n"
		"		<input type=\"submit\" value=\"Create\" class=\"button\"/>\n"
		"	</td>\n"
		"</tr>\n"
		"</table>\n"
		"</form>\n");
}

void
parse_query(void)
{
	char *query, *p;

	if (!(query = getenv("QUERY_STRING")))
		return;

	if ((p = getparam(query, "create"))) {
		if (decodeparam(create, sizeof(create), p) == -1) {
			http_status(401);
			exit(1);
		}
	}

	if ((p = getparam(query, "poll"))) {
		if (decodeparam(poll, sizeof(poll), p) == -1) {
			http_status(401);
			exit(1);
		}
	}

	if ((p = getparam(query, "question"))) {
		if (decodeparam(question, sizeof(question), p) == -1) {
			http_status(401);
			exit(1);
		}
	}

	if ((p = getparam(query, "options"))) {
		if (decodeparam(options, sizeof(options), p) == -1) {
			http_status(401);
			exit(1);
		}
	}

	if ((p = getparam(query, "choice"))) {
		if (decodeparam(choice, sizeof(choice), p) == -1) {
			http_status(401);
			exit(1);
		}
	}
}

int
main(void)
{
	size_t c;
	const char *errstr;
	struct stat sb;

	if (unveil(POLLS_DIR, "rwc") == -1) {
		http_status(500);
		err(1, "unveil");
	}

	if (pledge("stdio cpath rpath wpath", NULL) == -1) {
		http_status(500);
		err(1, "pledge");
	}

	if (stat(POLLS_DIR, &sb) == -1) {
		if (mkdir(POLLS_DIR, 0755) == -1) {
			http_status(500);
			err(1, "mkdir '%s' failed", POLLS_DIR);
		}
	}

	parse_query();

	if (*create) {
		if (create_poll_file(poll, question, options) == 0) {
			print_html_head();
			show_poll(poll, 0);
			print_html_foot();
		}
	} else if (*poll) {
		if (*choice) {
			c = strtonum(choice, 0, 256, &errstr);
			if (errstr != NULL)
				errx(1, "could not parse choice: %s, %s", errstr, choice);
			if (c > 0) {
				increment_option(poll, c);
				http_status(307);
			} else {
				print_html_head();
				show_poll(poll, 0);
				print_html_foot();
			}
		} else {
			print_html_head();
			show_poll(poll, 1);
			print_html_foot();
		}
	} else {
		print_html_head();
		list_polls();
		print_poll_create_form();
		print_html_foot();
	}

	return 0;
}
